function [ coef_ML, var_ML_nonrobust, var_ML_robust, sigma_ML ] = TransML( DATA, time_effect, ini_MDE )
 
% compute the transformed likelihood estimator by Hsiao et al 2002,
% for both AR(1) and ARX(1) models
 
% INPUT:
%  DATA  :      (T+1) x N x (K+1) matrix (K: number of exogenous variavles(X), K=0 corresponds to pure AR(1) case )
% time_effect:  0 -> a model without time effects,  1 -> a model with time effects
% ini_MDE:      0 -> random starting value for optimization,   1 -> Minimum distance estimator is used for optimization 

% OUTPUT:
% coef_ML:           coefficient of (gamma, beta)
% var_ML_nonrobust:  non-robust variance matrix of (gamma, beta) 
% var_ML_robust:     robust variance matrix of (gamma, beta)
% sigma_ML:          estimate of sigma^2


[TT, N, KK]  = size(DATA);
T = TT-1;
K = KK-1;
T1 = T-1;
if K==0;
    Y_NT = DATA(:,:,1); % T+1 x N
    Dy  = Y_NT(2:T+1,:) - Y_NT(1:T,:) ; % T x N (Dy_1,...,Dy_T)
    if time_effect==0;
        DW = zeros(T,N,2);
        for i=1:N
            DW(:,i,:) =  [1,   0;
                        zeros(T1,1), Dy(1:T1,i) ]; 
            
        end
    end    
    if time_effect==1;
        DW = zeros(T,N,2+T1);
        for i=1:N
            DW(:,i,:) = [1,   zeros(1,1+T1);
                        zeros(T1,1), Dy(1:T1,i) eye(T1) ];    
        end
    end    
end
if K>0;
    Y_NT = DATA(:,:,1); % T+1 x N
    X_NT = DATA(:,:,2:KK); % T+1 x N
    Dy  = Y_NT(2:T+1,:) - Y_NT(1:T,:) ; % T x N (Dy_1,...,Dy_T)
    DX  = X_NT(2:T+1,:,:) - X_NT(1:T,:,:); % T x N x K (DX_1,...,DX_T)
    if time_effect==0;
        DW = zeros(T,N,1+K*T+K+1);
        for i=1:N
            DW(:,i,:) = [1, reshape(squeeze(DX(1:T,i,:)),K*T,1)',  zeros(1,K+1);
                           zeros(T1,1+K*T), Dy(1:T1,i), squeeze(DX(2:T,i,:))   ];    
        end
    end
    if time_effect==1;
        DW = zeros(T,N,1+K*T+K+1+T1);
        for i=1:N
            DW(:,i,:) = [1, reshape(squeeze(DX(1:T,i,:)),K*T,1)',  zeros(1,K+1+T1);
                           zeros(T1,1+K*T), Dy(1:T1,i), squeeze(DX(2:T,i,:))  eye(T1) ];    
        end
    end
end


Omega = 2*eye(T)+diag(-1*ones(T-1,1),1)+diag(-1*ones(T-1,1),-1);
Phi = sort((1:T)','descend')*sort((1:T),'descend');
 



% compute starting values for MLE
% GMM using stacked IVs
D = [-eye(T-1) zeros(T-1,1)] + [zeros(T-1,1) eye(T-1) ];
DD = D*D';
if time_effect==0;
    if K==0;    m_ML = 3;     end
    if K>0;     m_ML = (K+1)*2;     end
end
if time_effect==1;
    if K==0;    m_ML = 3 + T1;     end
    if K>0;     m_ML = (K+1)*2 + T1;     end
end
Zy_ML = zeros(m_ML,1);  ZX_ML = zeros(m_ML,K+1+time_effect*T1);  ZZ_ML = zeros(m_ML,m_ML);

DX_GMM_N = zeros(T1,N,1+K+T1*time_effect);

for i=1:N;
    Dyi_GMM = Dy(2:T,i);   
    if time_effect==0
        if K==0;
        DXi_GMM = Dy(1:T-1,i);
        Zi_ML = [Y_NT(1:T-1,i) [zeros(1,1); Y_NT(1:T-2,i)] [zeros(2,1); Y_NT(1:T-3,i)] ];
        end

        if K>0;
        DXi_GMM = [Dy(1:T-1,i) squeeze(DX(2:T,i,:))];
        Zi_ML = [Y_NT(1:T-1,i) squeeze(X_NT(1:T-1,i,:)) [zeros(1,K+1); Y_NT(1:T-2,i) squeeze(X_NT(1:T-2,i,:))] ];   
        end
    end 

    if time_effect==1
        if K==0;
        DXi_GMM = [Dy(1:T-1,i) eye(T1)];
        Zi_ML = [Y_NT(1:T-1,i) [zeros(1,1); Y_NT(1:T-2,i)] [zeros(2,1); Y_NT(1:T-3,i)] eye(T1)];
        end

        if K>0;
        DXi_GMM = [Dy(1:T-1,i) squeeze(DX(2:T,i,:)) eye(T1)];
        Zi_ML = [Y_NT(1:T-1,i) squeeze(X_NT(1:T-1,i,:)) [zeros(1,K+1); Y_NT(1:T-2,i) squeeze(X_NT(1:T-2,i,:))] eye(T1) ];   
        end
    end 

    DX_GMM_N(:,i,:) = DXi_GMM;
    Zy_ML = Zy_ML + Zi_ML'*Dyi_GMM;   
    ZX_ML = ZX_ML + Zi_ML'*DXi_GMM;   
    ZZ_ML = ZZ_ML + Zi_ML'*DD*Zi_ML;
end;

delta_GMM0 = (ZX_ML'*inv(ZZ_ML)*ZX_ML)\(ZX_ML'*inv(ZZ_ML)*Zy_ML);

du = zeros(T1,N);
for i=1:N
  du(:,i) =  Dy(2:T,i) -  squeeze(DX_GMM_N(:,i,:))*delta_GMM0; 
end
% if time_effect==0
%     if K==0;  du = Dy(2:T,:) - Dy(1:T-1,:)*delta_GMM0(1) ;  end
%     if K==1;  du = Dy(2:T,:) - Dy(1:T-1,:)*delta_GMM0(1) - DX(2:T,:)*delta_GMM0(2);  end
% end
% 
% if time_effect==1
%     if K==0;  du = Dy(2:T,:) - Dy(1:T-1,:)*delta_GMM0(1)  ;  end
%     if K==1;  du = Dy(2:T,:) - Dy(1:T-1,:)*delta_GMM0(1) - DX(2:T,:)*delta_GMM0(2);  end
% end

sig2_u_ini = sum(sum(du.*du))/(2*N*(T-2));

% sig2_u_ini=1;

% initial estimate for phi
if K==0
    y1 = Dy(1,:)';
    X1 = ones(N,1);
end
if K>0 
    y1 = Dy(1,:)';
    X10 = zeros(N,T*K);
    for i=1:N
    X10(i,:) = vec(squeeze(DX(:,i,:))')';    
    end
    X1 = [ones(N,1), X10];
end

b_pi_ini = (X1'*X1)\(X1'*y1);

phi_ini = [b_pi_ini; delta_GMM0];
% v = y1-X1*b_pi_ini;
% MDE_omega = v'*v/(sig2_u_ini*(N-T-1));
% Omegastar(1,1) = MDE_omega;

% compute the minimum distance estimator

MDE_omega = (T-1)/T;
for i=1:N
    Dyi = Dy(:,i);
    DWi = squeeze(DW(:,i,:));
    ri =  Dyi - DWi*phi_ini;
    MDE_omega = MDE_omega + ri'*Phi*ri/(sig2_u_ini*N*T^2);
end
Omega(1,1) = MDE_omega;
invOmega = inv(Omega);
 
num=0;  den=0;
 for i=1:N
    Dyi = Dy(:,i);
    DWi = squeeze(DW(:,i,:));
     num = num + DWi'*invOmega*Dyi;
     den = den + DWi'*invOmega*DWi;
 end
MDE_phi = den\num;
 
MDE_sigma = 0;
 for i=1:N
    Dyi = Dy(:,i);
    DWi = squeeze(DW(:,i,:));
    ri =  Dyi - DWi*MDE_phi;
     MDE_sigma = MDE_sigma+ ri'*invOmega*ri/(N*T);
 end
 
% var_MDE_phi = mde_sigma*inv(den);


% correct MDE
if MDE_omega < ((T-1)/T);  MDE_omega = ((T-1)/T)+0.001; end;
if MDE_phi(K*T+2,1) > 1; MDE_phi(K*T+2,1) = 0.999; end
 


% Set the starting value for optimization of transML estimation

if ini_MDE==0;
    if time_effect==0
        phi_ini = [rand(1+K*T,1)*10-5; rand(1,1)*2-1; rand(K,1)*10-5];
    end
    if time_effect==1
        phi_ini = [rand(1+K*T,1)*10-5; rand(1,1)*2-1; rand(K,1)*10-5; rand(T1,1)*10-5; ];
    end
    omega_ini = (T-1)/T + rand(1,1)*10;
    sigma_ini =  rand(1,1)*10;
    theta_ini = [phi_ini; omega_ini; sigma_ini];
end

if ini_MDE==1;
    theta_ini = [MDE_phi;  MDE_omega;  MDE_sigma];
end;

if time_effect==0;
    if K==0
    lq = [ -inf;   -0.999;    (T-1)/T;  0.001 ];
    uq = [ inf;     0.999;       inf;     inf  ];
    end
    if K>0
    lq = [ -inf; -inf*ones(K*T,1);  -0.999;  -inf*ones(K,1);  (T-1)/T;  0.001 ];
    uq = [ inf;   inf*ones(K*T,1);   0.999;   inf*ones(K,1);    inf;     inf  ];
    end
end

if time_effect==1;
    if K==0
    lq = [ -inf;   -0.999;  -inf*ones(T-1,1);  (T-1)/T;  0.001 ];
    uq = [ inf;     0.999;   inf*ones(T-1,1);     inf;     inf  ];
    end
    if K>0
    lq = [ -inf; -inf*ones(K*T,1);  -0.999;  -inf*ones(K,1);  -inf*ones(T-1,1); (T-1)/T;  0.001 ];
    uq = [ inf;   inf*ones(K*T,1);   0.999;   inf*ones(K,1);   inf*ones(T-1,1);  inf;     inf  ];
    end
end


[theta, fval, exitflag, output, lambda, grad, hessian] = TransML_opt(N,T,K,Dy,DW,Phi,Omega,theta_ini,lq,uq);

dim_phi = size(DW,3);

phi_ML   = theta(1:dim_phi);
omega_ML = theta(dim_phi+1,1);
sigma_ML = theta(dim_phi+2,1);
Omega(1,1) = omega_ML;
invOmega = pinv(Omega);

sig2_uhat = sigma_ML;

% compute asymptotic variance
A11 = zeros(dim_phi,dim_phi);
A21 = zeros(1,dim_phi);

B11 = zeros(dim_phi,dim_phi);
B22 = zeros(1,1);
B33 = zeros(1,1);
B21 = zeros(1,dim_phi);
B31 = zeros(1,dim_phi);
B32 = zeros(1,1);

q_num=0;
q_den=0;

for i=1:N
    Dyi = Dy(:,i);
    DWi = squeeze(DW(:,i,:));
    ri =  Dyi - DWi*phi_ML;
    q_num = q_num + ri'*ri; 
    q_den = q_den + (ri'*invOmega*ri)^2; 

    A11 = A11 + DWi'*invOmega*DWi;
    A21 = A21 + ri'*Phi*DWi;
    B11 = B11 + DWi'*invOmega*(ri*ri')*invOmega*DWi;
    B22 = B22 + (ri'*Phi*ri/T)^2;
    B33 = B33 + (ri'*invOmega*ri/T)^2;
    B21 = B21 + (ri'*invOmega*DWi)*(ri'*Phi*ri);
    B31 = B31 + (ri'*invOmega*DWi)*(ri'*invOmega*ri);
    B32 = B32 + (ri'*Phi*ri/T)*(ri'*invOmega*ri/T);
    
end
g = 1+T*(omega_ML-1);


A11 = (1/(N*sig2_uhat))*A11;
A22 = T^2/(2*g^2);
A33 = T/(2*sig2_uhat^2);
A21 = (1/(g^2*N*sig2_uhat))*A21;
A12 = A21';
A31 = zeros(1,dim_phi);
A13 = A31';
A32 = T/(2*g*sig2_uhat);
A23 = A32';
A = [A11 A12 A13; 
     A21 A22 A23; 
     A31 A32 A33];

B11 = (1/(N*sig2_uhat^2))*B11;
B22 = (T^2/(4*g^4*sig2_uhat^2))*((1/N)*B22 - g^2*sig2_uhat^2);
B33 = (T^2/(4*sig2_uhat^4))*((1/N)*B33 - sig2_uhat^2);
B21 = (1/(2*N*g^2*sig2_uhat^2))*B21;
B31 = (1/(2*N*sig2_uhat^3))*B31;
B32 = (T^2/(4*g^2*sig2_uhat^3))*((1/N)*B32 - g*sig2_uhat^2);
B12 = B21';
B13 = B31';
B23 = B32';
B = [B11 B12 B13; 
     B21 B22 B23;
     B31 B32 B33];

%    disp([(A)]);
%    disp([B]);

var_theta_nonrobust = pinv(A)/(N);
var_theta_robust = pinv(A)*B*pinv(A)/(N);
% var_theta_robust2 = pinv(A)*B_2*pinv(A)/(N*T);

%  disp(N*[var_theta_nonrobust var_theta_robust]);


% disp([var_theta_robust1 var_theta_robust2]);

coef_ML    = theta(K*T+2:K*T+K+2,1);
var_ML_nonrobust = var_theta_nonrobust(K*T+2:K*T+K+2, K*T+2:K*T+K+2);
var_ML_robust    = var_theta_robust(K*T+2:K*T+K+2, K*T+2:K*T+K+2);
% var_ML_robust2 = var_theta_robust2(K*T+2:K*T+K+2, K*T+2:K*T+K+2);

end

